# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

from sphinx.ext import autodoc
from sphinx.ext.autodoc import *
from sphinx.util import force_decode
from sphinx.util.docstrings import prepare_docstring
import inspect

# stop warnings cluttering the make output
import warnings
warnings.filterwarnings("ignore")


class ClassWithConstructorDocumenter(autodoc.ClassDocumenter):
    priority = 1000000

    def get_object_members(self, want_all):
        return autodoc.ClassDocumenter.get_object_members(self, want_all)

    @staticmethod
    def can_document_member(member, mname, isattr, self):
        return autodoc.ClassDocumenter.can_document_member(member, mname,
                                                           isattr, self)

    def get_doc(self, encoding=None):
        content = self.env.config.autoclass_content

        docstrings = []
        docstring = self.get_attr(self.object, '__doc__', None)
        if docstring:
            docstrings.append(docstring)

        # for classes, what the "docstring" is can be controlled via a
        # config value; the default is only the class docstring
        if content in ('both', 'init'):
            constructor = self.get_constructor()
            if constructor:
                initdocstring = self.get_attr(constructor, '__doc__', None)
            else:
                initdocstring = None
            if initdocstring:
                if content == 'init':
                    docstrings = [initdocstring]
                else:
                    docstrings.append(initdocstring)

        return [prepare_docstring(force_decode(docstring, encoding))
                for docstring in docstrings]

    def get_constructor(self):
        # for classes, the relevant signature is the __init__ method's
        initmeth = self.get_attr(self.object, '__new__', None)

        if initmeth is None or initmeth is object.__new__ or not \
                (inspect.ismethod(initmeth) or inspect.isfunction(initmeth)):
            initmeth = None

        if initmeth is None:
            initmeth = self.get_attr(self.object, '__init__', None)

        if initmeth is None or initmeth is object.__init__ or \
                initmeth is object.__new__ or not \
                (inspect.ismethod(initmeth) or inspect.isfunction(initmeth)):
            initmeth = None

        return initmeth

    def format_args(self):
        initmeth = self.get_constructor()
        try:
            argspec = inspect.getargspec(initmeth)
        except TypeError:
            # still not possible: happens e.g. for old-style classes
            # with __init__ in C
            return None
        if argspec[0] and argspec[0][0] in ('cls', 'self'):
            del argspec[0][0]
        return inspect.formatargspec(*argspec)


def setup(app):
    app.add_autodocumenter(ClassWithConstructorDocumenter, override=True)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function :func:\
`iris.fileformats.um._fast_load_structured_fields.group_structured_fields`.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

from unittest import mock

from iris.fileformats.um._fast_load_structured_fields import (
    group_structured_fields,
)


def _convert_to_vector(value, length, default):
    """
    Return argument (or default) in a list of length 'length'.

    The 'value' arg must either be scalar, or a list of length 'length'.
    A value of None is replaced by the default.
    If scalar, the value is replicated to the required length.

    """
    if value is None:
        value = default
    if hasattr(value, "__len__"):
        assert len(value) == length
    else:
        value = [value] * length
    return value


class Test__grouping(tests.IrisTest):
    def _dummy_fields_iter(self, stashes=None, models=None, lbprocs=None):
        # Make a group of test fields, and return an iterator over it.
        a_vec = [vec for vec in (stashes, models, lbprocs) if vec is not None]
        number = len(a_vec[0])
        stashes = _convert_to_vector(stashes, number, default=31)
        models = _convert_to_vector(models, number, default=71)
        lbprocs = _convert_to_vector(lbprocs, number, default=91)
        self.test_fields = [
            mock.MagicMock(
                lbuser=[0, 0, 0, x_stash, 0, 0, x_model],
                lbproc=x_lbproc,
                i_field=ind + 1001,
            )
            for ind, x_stash, x_model, x_lbproc in zip(
                range(number), stashes, models, lbprocs
            )
        ]
        return (fld for fld in self.test_fields)

    def _group_result(self, fields):
        # Run the testee, but returning just the groups (not FieldCollations).
        result = list(group_structured_fields(fields, collation_class=tuple))
        return result

    def _test_fields(self, item):
        # Convert nested tuples/lists of field-numbers into fields.
        if isinstance(item, int):
            result = self.test_fields[item - 1001]
        else:
            result = type(item)(self._test_fields(el) for el in item)
        return result

    def test_none(self):
        null_iter = (x for x in [])
        result = self._group_result(null_iter)
        self.assertEqual(result, [])

    def test_one(self):
        fields_iter = self._dummy_fields_iter(stashes=[1])
        result = self._group_result(fields_iter)
        self.assertEqual(result, self._test_fields([(1001,)]))

    def test_allsame(self):
        fields_iter = self._dummy_fields_iter(stashes=[1, 1, 1])
        result = self._group_result(fields_iter)
        self.assertEqual(result, self._test_fields([(1001, 1002, 1003)]))

    def test_stashes_different(self):
        fields_iter = self._dummy_fields_iter(stashes=[1, 1, 22, 1, 22, 333])
        result = self._group_result(fields_iter)
        self.assertEqual(
            result,
            self._test_fields([(1001, 1002, 1004), (1003, 1005), (1006,)]),
        )

    def test_models_different(self):
        fields_iter = self._dummy_fields_iter(models=[10, 21, 10])
        result = self._group_result(fields_iter)
        self.assertEqual(result, self._test_fields([(1001, 1003), (1002,)]))

    def test_lbprocs_different(self):
        fields_iter = self._dummy_fields_iter(lbprocs=[991, 995, 991])
        result = self._group_result(fields_iter)
        self.assertEqual(result, self._test_fields([(1001, 1003), (1002,)]))

    def test_2d_combines(self):
        fields_iter = self._dummy_fields_iter(
            stashes=[11, 11, 15, 11], lbprocs=[31, 42, 31, 42]
        )
        result = self._group_result(fields_iter)
        self.assertEqual(
            result, self._test_fields([(1001,), (1002, 1004), (1003,)])
        )

    def test_sortorder(self):
        fields_iter = self._dummy_fields_iter(stashes=[11, 7, 12])
        result = self._group_result(fields_iter)
        self.assertEqual(
            result, self._test_fields([(1002,), (1001,), (1003,)])
        )

    def test_sortorder_2d(self):
        fields_iter = self._dummy_fields_iter(
            stashes=[11, 11, 12], lbprocs=[31, 9, 1]
        )
        result = self._group_result(fields_iter)
        self.assertEqual(
            result, self._test_fields([(1002,), (1001,), (1003,)])
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import copy
import functools
import warnings

import numpy as np
import numpy.ma as ma

from iris._lazy_data import map_complete_blocks
from iris.analysis._interpolation import (
    EXTRAPOLATION_MODES,
    extend_circular_coord_and_data,
    get_xy_dim_coords,
    snapshot_grid,
)
from iris.analysis._scipy_interpolate import _RegularGridInterpolator
import iris.cube
from iris.util import _meshgrid

from scipy.sparse import csc_matrix, diags as sparse_diags


def _transform_xy_arrays(crs_from, x, y, crs_to):
    """
    Transform 2d points between cartopy coordinate reference systems.

    NOTE: copied private function from iris.analysis.cartography.

    Args:

    * crs_from, crs_to (:class:`cartopy.crs.Projection`):
        The coordinate reference systems.
    * x, y (arrays):
        point locations defined in 'crs_from'.

    Returns:
        x, y :  Arrays of locations defined in 'crs_to'.

    """
    pts = crs_to.transform_points(crs_from, x, y)
    return pts[..., 0], pts[..., 1]


def _regrid_weighted_curvilinear_to_rectilinear__prepare(
    src_cube, weights, grid_cube
):
    """
    First (setup) part of 'regrid_weighted_curvilinear_to_rectilinear'.

    Check inputs and calculate the sparse regrid matrix and related info.
    The 'regrid info' returned can be re-used over many 2d slices.

    """
    if src_cube.aux_factories:
        msg = "All source cube derived coordinates will be ignored."
        warnings.warn(msg)

    # Get the source cube x and y 2D auxiliary coordinates.
    sx, sy = src_cube.coord(axis="x"), src_cube.coord(axis="y")
    # Get the target grid cube x and y dimension coordinates.
    tx, ty = get_xy_dim_coords(grid_cube)

    if sx.units != sy.units:
        msg = (
            "The source cube x ({!r}) and y ({!r}) coordinates must "
            "have the same units."
        )
        raise ValueError(msg.format(sx.name(), sy.name()))

    if src_cube.coord_dims(sx) != src_cube.coord_dims(sy):
        msg = (
            "The source cube x ({!r}) and y ({!r}) coordinates must "
            "map onto the same cube dimensions."
        )
        raise ValueError(msg.format(sx.name(), sy.name()))

    if sx.coord_system != sy.coord_system:
        msg = (
            "The source cube x ({!r}) and y ({!r}) coordinates must "
            "have the same coordinate system."
        )
        raise ValueError(msg.format(sx.name(), sy.name()))

    if sx.coord_system is None:
        msg = (
            "The source X and Y coordinates must have a defined "
            "coordinate system."
        )
        raise ValueError(msg)

    if tx.units != ty.units:
        msg = (
            "The target grid cube x ({!r}) and y ({!r}) coordinates must "
            "have the same units."
        )
        raise ValueError(msg.format(tx.name(), ty.name()))

    if tx.coord_system is None:
        msg = (
            "The target X and Y coordinates must have a defined "
            "coordinate system."
        )
        raise ValueError(msg)

    if tx.coord_system != ty.coord_system:
        msg = (
            "The target grid cube x ({!r}) and y ({!r}) coordinates must "
            "have the same coordinate system."
        )
        raise ValueError(msg.format(tx.name(), ty.name()))

    if weights is None:
        weights = np.ones(sx.shape)
    if weights.shape != sx.shape:
        msg = (
            "Provided weights must have the same shape as the X and Y "
            "coordinates."
        )
        raise ValueError(msg)

    if not tx.has_bounds() or not tx.is_contiguous():
        msg = (
            "The target grid cube x ({!r})coordinate requires "
            "contiguous bounds."
        )
        raise ValueError(msg.format(tx.name()))

    if not ty.has_bounds() or not ty.is_contiguous():
        msg = (
            "The target grid cube y ({!r}) coordinate requires "
            "contiguous bounds."
        )
        raise ValueError(msg.format(ty.name()))

    def _src_align_and_flatten(coord):
        # Return a flattened, unmasked copy of a coordinate's points array that
        # will align with a flattened version of the source cube's data.
        #
        # PP-TODO: Should work with any cube dimensions for X and Y coords.
        #  Probably needs fixing anyway?
        #
        points = coord.points
        if src_cube.coord_dims(coord) == (1, 0):
            points = points.T
        if points.shape != src_cube.shape:
            msg = (
                "The shape of the points array of {!r} is not compatible "
                "with the shape of {!r}."
            )
            raise ValueError(msg.format(coord.name(), src_cube.name()))
        return np.asarray(points.flatten())

    # Align and flatten the coordinate points of the source space.
    sx_points = _src_align_and_flatten(sx)
    sy_points = _src_align_and_flatten(sy)

    # Transform source X and Y points into the target coord-system, if needed.
    if sx.coord_system != tx.coord_system:
        src_crs = sx.coord_system.as_cartopy_projection()
        tgt_crs = tx.coord_system.as_cartopy_projection()
        sx_points, sy_points = _transform_xy_arrays(
            src_crs, sx_points, sy_points, tgt_crs
        )
    #
    # TODO: how does this work with scaled units ??
    #  e.g. if crs is latlon, units could be degrees OR radians ?
    #

    # Wrap modular values (e.g. longitudes) if required.
    modulus = sx.units.modulus
    if modulus is not None:
        # Match the source cube x coordinate range to the target grid
        # cube x coordinate range.
        min_sx, min_tx = np.min(sx.points), np.min(tx.points)
        if min_sx < 0 and min_tx >= 0:
            indices = np.where(sx_points < 0)
            # Ensure += doesn't raise a TypeError
            if not np.can_cast(modulus, sx_points.dtype):
                sx_points = sx_points.astype(type(modulus), casting="safe")
            sx_points[indices] += modulus
        elif min_sx >= 0 and min_tx < 0:
            indices = np.where(sx_points > (modulus / 2))
            # Ensure -= doesn't raise a TypeError
            if not np.can_cast(modulus, sx_points.dtype):
                sx_points = sx_points.astype(type(modulus), casting="safe")
            sx_points[indices] -= modulus

    # Create target grid cube x and y cell boundaries.
    tx_depth, ty_depth = tx.points.size, ty.points.size
    (tx_dim,) = grid_cube.coord_dims(tx)
    (ty_dim,) = grid_cube.coord_dims(ty)

    tx_cells = np.concatenate((tx.bounds[:, 0], tx.bounds[-1, 1].reshape(1)))
    ty_cells = np.concatenate((ty.bounds[:, 0], ty.bounds[-1, 1].reshape(1)))

    # Determine the target grid cube x and y cells that bound
    # the source cube x and y points.

    def _regrid_indices(cells, depth, points):
        # Calculate the minimum difference in cell extent.
        extent = np.min(np.diff(cells))
        if extent == 0:
            # Detected an dimension coordinate with an invalid
            # zero length cell extent.
            msg = (
                "The target grid cube {} ({!r}) coordinate contains "
                "a zero length cell extent."
            )
            axis, name = "x", tx.name()
            if points is sy_points:
                axis, name = "y", ty.name()
            raise ValueError(msg.format(axis, name))
        elif extent > 0:
            # The cells of the dimension coordinate are in ascending order.
            indices = np.searchsorted(cells, points, side="right") - 1
        else:
            # The cells of the dimension coordinate are in descending order.
            # np.searchsorted() requires ascending order, so we require to
            # account for this restriction.
            cells = cells[::-1]
            right = np.searchsorted(cells, points, side="right")
            left = np.searchsorted(cells, points, side="left")
            indices = depth - right
            # Only those points that exactly match the left-hand cell bound
            # will differ between 'left' and 'right'. Thus their appropriate
            # target cell location requires to be recalculated to give the
            # correct descending [upper, lower) interval cell, source to target
            # regrid behaviour.
            delta = np.where(left != right)[0]
            if delta.size:
                indices[delta] = depth - left[delta]
        return indices

    x_indices = _regrid_indices(tx_cells, tx_depth, sx_points)
    y_indices = _regrid_indices(ty_cells, ty_depth, sy_points)

    # Now construct a sparse M x N matix, where M is the flattened target
    # space, and N is the flattened source space. The sparse matrix will then
    # be populated with those source cube points that contribute to a specific
    # target cube cell.

    # Determine the valid indices and their offsets in M x N space.
    # Calculate the valid M offsets.
    cols = np.where(
        (y_indices >= 0)
        & (y_indices < ty_depth)
        & (x_indices >= 0)
        & (x_indices < tx_depth)
    )[0]

    # Reduce the indices to only those that are valid.
    x_indices = x_indices[cols]
    y_indices = y_indices[cols]

    # Calculate the valid N offsets.
    if ty_dim < tx_dim:
        rows = y_indices * tx.points.size + x_indices
    else:
        rows = x_indices * ty.points.size + y_indices

    # Calculate the associated valid weights.
    weights_flat = weights.flatten()
    data = weights_flat[cols]

    # Build our sparse M x N matrix of weights.
    sparse_matrix = csc_matrix(
        (data, (rows, cols)), shape=(grid_cube.data.size, src_cube.data.size)
    )

    # Performing a sparse sum to collapse the matrix to (M, 1).
    sum_weights = sparse_matrix.sum(axis=1).getA()

    # Determine the rows (flattened target indices) that have a
    # contribution from one or more source points.
    rows = np.nonzero(sum_weights)

    # NOTE: when source points are masked, this 'sum_weights' is possibly
    # incorrect and needs re-calculating.  Likewise 'rows' may cover target
    # cells which happen to get no data.  This is dealt with by adjusting as
    # required in the '__perform' function, below.

    regrid_info = (sparse_matrix, sum_weights, rows, grid_cube)
    return regrid_info


def _regrid_weighted_curvilinear_to_rectilinear__perform(
    src_cube, regrid_info
):
    """
    Second (regrid) part of 'regrid_weighted_curvilinear_to_rectilinear'.

    Perform the prepared regrid calculation on a single 2d cube.

    """
    sparse_matrix, sum_weights, rows, grid_cube = regrid_info

    # Calculate the numerator of the weighted mean (M, 1).
    is_masked = ma.isMaskedArray(src_cube.data)
    if not is_masked:
        data = src_cube.data
    else:
        # Use raw data array
        data = src_cube.data.data
        # Check if there are any masked source points to take account of.
        is_masked = np.ma.is_masked(src_cube.data)
        if is_masked:
            # Zero any masked source points so they add nothing in output sums.
            mask = src_cube.data.mask
            data[mask] = 0.0
            # Calculate a new 'sum_weights' to allow for missing source points.
            # N.B. it is more efficient to use the original once-calculated
            # sparse matrix, but in this case we can't.
            # Hopefully, this post-multiplying by the validities is less costly
            # than repeating the whole sparse calculation.
            valid_src_cells = ~mask.flat[:]
            src_cell_validity_factors = sparse_diags(
                np.array(valid_src_cells, dtype=int), 0
            )
            valid_weights = sparse_matrix * src_cell_validity_factors
            sum_weights = valid_weights.sum(axis=1).getA()
            # Work out where output cells are missing all contributions.
            # This allows for where 'rows' contains output cells that have no
            # data because of missing input points.
            zero_sums = sum_weights == 0.0
            # Make sure we can still divide by sum_weights[rows].
            sum_weights[zero_sums] = 1.0

    # Calculate sum in each target cell, over contributions from each source
    # cell.
    numerator = sparse_matrix * data.reshape(-1, 1)

    # Create a template for the weighted mean result.
    weighted_mean = ma.masked_all(numerator.shape, dtype=numerator.dtype)

    # Calculate final results in all relevant places.
    weighted_mean[rows] = numerator[rows] / sum_weights[rows]
    if is_masked:
        # Ensure masked points where relevant source cells were all missing.
        if np.any(zero_sums):
            # Make masked if it wasn't.
            weighted_mean = np.ma.asarray(weighted_mean)
            # Mask where contributing sums were zero.
            weighted_mean[zero_sums] = np.ma.masked

    # Construct the final regridded weighted mean cube.
    tx = grid_cube.coord(axis="x", dim_coords=True)
    ty = grid_cube.coord(axis="y", dim_coords=True)
    (tx_dim,) = grid_cube.coord_dims(tx)
    (ty_dim,) = grid_cube.coord_dims(ty)
    dim_coords_and_dims = list(zip((ty.copy(), tx.copy()), (ty_dim, tx_dim)))
    cube = iris.cube.Cube(
        weighted_mean.reshape(grid_cube.shape),
        dim_coords_and_dims=dim_coords_and_dims,
    )
    cube.metadata = copy.deepcopy(src_cube.metadata)

    for coord in src_cube.coords(dimensions=()):
        cube.add_aux_coord(coord.copy())

    return cube


class CurvilinearRegridder:
    """
    This class provides support for performing point-in-cell regridding
    between a curvilinear source grid and a rectilinear target grid.

    """

    def __init__(self, src_grid_cube, target_grid_cube, weights=None):
        """
        Create a regridder for conversions between the source
        and target grids.

        Args:

        * src_grid_cube:
            The :class:`~iris.cube.Cube` providing the source grid.
        * tgt_grid_cube:
            The :class:`~iris.cube.Cube` providing the target grid.

        Optional Args:

        * weights:
            A :class:`numpy.ndarray` instance that defines the weights
            for the grid cells of the source grid. Must have the same shape
            as the data of the source grid.
            If unspecified, equal weighting is assumed.

        """
        # Validity checks.
        if not isinstance(src_grid_cube, iris.cube.Cube):
            raise TypeError("'src_grid_cube' must be a Cube")
        if not isinstance(target_grid_cube, iris.cube.Cube):
            raise TypeError("'target_grid_cube' must be a Cube")
        # Snapshot the state of the cubes to ensure that the regridder
        # is impervious to external changes to the original source cubes.
        self._src_cube = src_grid_cube.copy()
        self._target_cube = target_grid_cube.copy()
        self.weights = weights
        self._regrid_info = None

    @staticmethod
    def _get_horizontal_coord(cube, axis):
        """
        Gets the horizontal coordinate on the supplied cube along the
        specified axis.

        Args:

        * cube:
            An instance of :class:`iris.cube.Cube`.
        * axis:
            Locate coordinates on `cube` along this axis.

        Returns:
            The horizontal coordinate on the specified axis of the supplied
            cube.

        """
        coords = cube.coords(axis=axis, dim_coords=False)
        if len(coords) != 1:
            raise ValueError(
                "Cube {!r} must contain a single 1D {} "
                "coordinate.".format(cube.name(), axis)
            )
        return coords[0]

    def __call__(self, src):
        """
        Regrid the supplied :class:`~iris.cube.Cube` on to the target grid of
        this :class:`_CurvilinearRegridder`.

        The given cube must be defined with the same grid as the source
        grid used to create this :class:`_CurvilinearRegridder`.

        If the source cube has lazy data, it will be realized before
        regridding and the returned cube will also have realized data.

        Args:

        * src:
            A :class:`~iris.cube.Cube` to be regridded.

        Returns:
            A cube defined with the horizontal dimensions of the target
            and the other dimensions from this cube. The data values of
            this cube will be converted to values on the new grid using
            point-in-cell regridding.

        """
        # Validity checks.
        if not isinstance(src, iris.cube.Cube):
            raise TypeError("'src' must be a Cube")

        gx = self._get_horizontal_coord(self._src_cube, "x")
        gy = self._get_horizontal_coord(self._src_cube, "y")
        src_grid = (gx.copy(), gy.copy())
        sx = self._get_horizontal_coord(src, "x")
        sy = self._get_horizontal_coord(src, "y")
        if (sx, sy) != src_grid:
            raise ValueError(
                "The given cube is not defined on the same "
                "source grid as this regridder."
            )

        # Call the regridder function.
        # This includes repeating over any non-XY dimensions, because the
        # underlying routine does not support this.
        # FOR NOW: we will use cube.slices and merge to achieve this,
        # though that is not a terribly efficient method ...
        # TODO: create a template result cube and paste data slices into it,
        # which would be more efficient.
        result_slices = iris.cube.CubeList([])
        for slice_cube in src.slices(sx):
            if self._regrid_info is None:
                # Calculate the basic regrid info just once.
                self._regrid_info = (
                    _regrid_weighted_curvilinear_to_rectilinear__prepare(
                        slice_cube, self.weights, self._target_cube
                    )
                )
            slice_result = (
                _regrid_weighted_curvilinear_to_rectilinear__perform(
                    slice_cube, self._regrid_info
                )
            )
            result_slices.append(slice_result)
        result = result_slices.merge_cube()
        return result


class RectilinearRegridder:
    """
    This class provides support for performing nearest-neighbour or
    linear regridding between source and target grids.

    """

    def __init__(
        self, src_grid_cube, tgt_grid_cube, method, extrapolation_mode
    ):
        """
        Create a regridder for conversions between the source
        and target grids.

        Args:

        * src_grid_cube:
            The :class:`~iris.cube.Cube` providing the source grid.
        * tgt_grid_cube:
            The :class:`~iris.cube.Cube` providing the target grid.
        * method:
            Either 'linear' or 'nearest'.
        * extrapolation_mode:
            Must be one of the following strings:

              * 'extrapolate' - The extrapolation points will be
                calculated by extending the gradient of the closest two
                points.
              * 'nan' - The extrapolation points will be be set to NaN.
              * 'error' - An exception will be raised, notifying an
                attempt to extrapolate.
              * 'mask' - The extrapolation points will always be masked, even
                if the source data is not a MaskedArray.
              * 'nanmask' - If the source data is a MaskedArray the
                extrapolation points will be masked. Otherwise they will be
                set to NaN.

        """
        # Validity checks.
        if not isinstance(src_grid_cube, iris.cube.Cube):
            raise TypeError("'src_grid_cube' must be a Cube")
        if not isinstance(tgt_grid_cube, iris.cube.Cube):
            raise TypeError("'tgt_grid_cube' must be a Cube")
        # Snapshot the state of the cubes to ensure that the regridder
        # is impervious to external changes to the original source cubes.
        self._src_grid = snapshot_grid(src_grid_cube)
        self._tgt_grid = snapshot_grid(tgt_grid_cube)
        # Check the target grid units.
        for coord in self._tgt_grid:
            self._check_units(coord)
        # Whether to use linear or nearest-neighbour interpolation.
        if method not in ("linear", "nearest"):
            msg = "Regridding method {!r} not supported.".format(method)
            raise ValueError(msg)
        self._method = method
        # The extrapolation mode.
        if extrapolation_mode not in EXTRAPOLATION_MODES:
            msg = "Invalid extrapolation mode {!r}"
            raise ValueError(msg.format(extrapolation_mode))
        self._extrapolation_mode = extrapolation_mode

    @property
    def method(self):
        return self._method

    @property
    def extrapolation_mode(self):
        return self._extrapolation_mode

    @staticmethod
    def _sample_grid(src_coord_system, grid_x_coord, grid_y_coord):
        """
        Convert the rectilinear grid coordinates to a curvilinear grid in
        the source coordinate system.

        The `grid_x_coord` and `grid_y_coord` must share a common coordinate
        system.

        Args:

        * src_coord_system:
            The :class:`iris.coord_system.CoordSystem` for the grid of the
            source Cube.
        * grid_x_coord:
            The :class:`iris.coords.DimCoord` for the X coordinate.
        * grid_y_coord:
            The :class:`iris.coords.DimCoord` for the Y coordinate.

        Returns:
            A tuple of the X and Y coordinate values as 2-dimensional
            arrays.

        """
        grid_x, grid_y = _meshgrid(grid_x_coord.points, grid_y_coord.points)
        # Skip the CRS transform if we can to avoid precision problems.
        if src_coord_system == grid_x_coord.coord_system:
            sample_grid_x = grid_x
            sample_grid_y = grid_y
        else:
            src_crs = src_coord_system.as_cartopy_crs()
            grid_crs = grid_x_coord.coord_system.as_cartopy_crs()
            sample_xyz = src_crs.transform_points(grid_crs, grid_x, grid_y)
            sample_grid_x = sample_xyz[..., 0]
            sample_grid_y = sample_xyz[..., 1]
        return sample_grid_x, sample_grid_y

    @staticmethod
    def _regrid(
        src_data,
        x_dim,
        y_dim,
        src_x_coord,
        src_y_coord,
        sample_grid_x,
        sample_grid_y,
        method="linear",
        extrapolation_mode="nanmask",
    ):
        """
        Regrid the given data from the src grid to the sample grid.

        The result will be a MaskedArray if either/both of:
         - the source array is a MaskedArray,
         - the extrapolation_mode is 'mask' and the result requires
           extrapolation.

        If the result is a MaskedArray the mask for each element will be set
        if either/both of:
         - there is a non-zero contribution from masked items in the input data
         - the element requires extrapolation and the extrapolation_mode
           dictates a masked value.

        Args:

        * src_data:
            An N-dimensional NumPy array or MaskedArray.
        * x_dim:
            The X dimension within `src_data`.
        * y_dim:
            The Y dimension within `src_data`.
        * src_x_coord:
            The X :class:`iris.coords.DimCoord`.
        * src_y_coord:
            The Y :class:`iris.coords.DimCoord`.
        * sample_grid_x:
            A 2-dimensional array of sample X values.
        * sample_grid_y:
            A 2-dimensional array of sample Y values.

        Kwargs:

        * method:
            Either 'linear' or 'nearest'. The default method is 'linear'.
        * extrapolation_mode:
            Must be one of the following strings:

              * 'linear' - The extrapolation points will be calculated by
                extending the gradient of the closest two points.
              * 'nan' - The extrapolation points will be be set to NaN.
              * 'error' - A ValueError exception will be raised, notifying an
                attempt to extrapolate.
              * 'mask' - The extrapolation points will always be masked, even
                if the source data is not a MaskedArray.
              * 'nanmask' - If the source data is a MaskedArray the
                extrapolation points will be masked. Otherwise they will be
                set to NaN.

            The default mode of extrapolation is 'nanmask'.

        Returns:
            The regridded data as an N-dimensional NumPy array. The lengths
            of the X and Y dimensions will now match those of the sample
            grid.

        """
        #
        # XXX: At the moment requires to be a static method as used by
        # experimental regrid_area_weighted_rectilinear_src_and_grid
        #
        if sample_grid_x.shape != sample_grid_y.shape:
            raise ValueError("Inconsistent sample grid shapes.")
        if sample_grid_x.ndim != 2:
            raise ValueError("Sample grid must be 2-dimensional.")

        # Prepare the result data array
        shape = list(src_data.shape)
        assert shape[x_dim] == src_x_coord.shape[0]
        assert shape[y_dim] == src_y_coord.shape[0]

        shape[y_dim] = sample_grid_x.shape[0]
        shape[x_dim] = sample_grid_x.shape[1]

        dtype = src_data.dtype
        if method == "linear":
            # If we're given integer values, convert them to the smallest
            # possible float dtype that can accurately preserve the values.
            if dtype.kind == "i":
                dtype = np.promote_types(dtype, np.float16)

        if ma.isMaskedArray(src_data):
            data = ma.empty(shape, dtype=dtype)
            data.mask = np.zeros(data.shape, dtype=np.bool)
        else:
            data = np.empty(shape, dtype=dtype)

        # The interpolation class requires monotonically increasing
        # coordinates, so flip the coordinate(s) and data if they aren't.
        reverse_x = (
            src_x_coord.points[0] > src_x_coord.points[1]
            if src_x_coord.points.size > 1
            else False
        )
        reverse_y = src_y_coord.points[0] > src_y_coord.points[1]
        flip_index = [slice(None)] * src_data.ndim
        if reverse_x:
            src_x_coord = src_x_coord[::-1]
            flip_index[x_dim] = slice(None, None, -1)
        if reverse_y:
            src_y_coord = src_y_coord[::-1]
            flip_index[y_dim] = slice(None, None, -1)
        src_data = src_data[tuple(flip_index)]

        if src_x_coord.circular:
            x_points, src_data = extend_circular_coord_and_data(
                src_x_coord, src_data, x_dim
            )
        else:
            x_points = src_x_coord.points

        # Slice out the first full 2D piece of data for construction of the
        # interpolator.
        index = [0] * src_data.ndim
        index[x_dim] = index[y_dim] = slice(None)
        initial_data = src_data[tuple(index)]
        if y_dim < x_dim:
            initial_data = initial_data.T

        # Construct the interpolator, we will fill in any values out of bounds
        # manually.
        interpolator = _RegularGridInterpolator(
            [x_points, src_y_coord.points],
            initial_data,
            method=method,
            bounds_error=False,
            fill_value=None,
        )
        # The constructor of the _RegularGridInterpolator class does
        # some unnecessary checks on these values, so we set them
        # afterwards instead. Sneaky. ;-)
        try:
            mode = EXTRAPOLATION_MODES[extrapolation_mode]
        except KeyError:
            raise ValueError("Invalid extrapolation mode.")
        interpolator.bounds_error = mode.bounds_error
        interpolator.fill_value = mode.fill_value

        # Construct the target coordinate points array, suitable for passing to
        # the interpolator multiple times.
        interp_coords = [
            sample_grid_x.astype(np.float64)[..., np.newaxis],
            sample_grid_y.astype(np.float64)[..., np.newaxis],
        ]

        # Map all the requested values into the range of the source
        # data (centred over the centre of the source data to allow
        # extrapolation where required).
        min_x, max_x = x_points.min(), x_points.max()
        if src_x_coord.units.modulus:
            modulus = src_x_coord.units.modulus
            offset = (max_x + min_x - modulus) * 0.5
            interp_coords[0] -= offset
            interp_coords[0] = (interp_coords[0] % modulus) + offset

        interp_coords = np.dstack(interp_coords)

        weights = interpolator.compute_interp_weights(interp_coords)

        def interpolate(data):
            # Update the interpolator for this data slice.
            data = data.astype(interpolator.values.dtype)
            if y_dim < x_dim:
                data = data.T
            interpolator.values = data
            data = interpolator.interp_using_pre_computed_weights(weights)
            if y_dim > x_dim:
                data = data.T
            return data

        # Build up a shape suitable for passing to ndindex, inside the loop we
        # will insert slice(None) on the data indices.
        iter_shape = list(shape)
        iter_shape[x_dim] = iter_shape[y_dim] = 1

        # Iterate through each 2d slice of the data, updating the interpolator
        # with the new data as we go.
        for index in np.ndindex(tuple(iter_shape)):
            index = list(index)
            index[x_dim] = index[y_dim] = slice(None)

            src_subset = src_data[tuple(index)]
            interpolator.fill_value = mode.fill_value
            data[tuple(index)] = interpolate(src_subset)

            if ma.isMaskedArray(data) or mode.force_mask:
                # NB. np.ma.getmaskarray returns an array of `False` if
                # `src_subset` is not a masked array.
                src_mask = np.ma.getmaskarray(src_subset)
                interpolator.fill_value = mode.mask_fill_value
                mask_fraction = interpolate(src_mask)
                new_mask = mask_fraction > 0

                if np.ma.isMaskedArray(data):
                    data.mask[tuple(index)] = new_mask
                elif np.any(new_mask):
                    # Set mask=False to ensure we have an expanded mask array.
                    data = np.ma.MaskedArray(data, mask=False)
                    data.mask[tuple(index)] = new_mask

        return data

    @staticmethod
    def _create_cube(
        data,
        src,
        x_dim,
        y_dim,
        src_x_coord,
        src_y_coord,
        grid_x_coord,
        grid_y_coord,
        sample_grid_x,
        sample_grid_y,
        regrid_callback,
    ):
        """
        Return a new Cube for the result of regridding the source Cube onto
        the new grid.

        All the metadata and coordinates of the result Cube are copied from
        the source Cube, with two exceptions:
            - Grid dimension coordinates are copied from the grid Cube.
            - Auxiliary coordinates which span the grid dimensions are
              ignored, except where they provide a reference surface for an
              :class:`iris.aux_factory.AuxCoordFactory`.

        Args:

        * data:
            The regridded data as an N-dimensional NumPy array.
        * src:
            The source Cube.
        * x_dim:
            The X dimension within the source Cube.
        * y_dim:
            The Y dimension within the source Cube.
        * src_x_coord:
            The X :class:`iris.coords.DimCoord`.
        * src_y_coord:
            The Y :class:`iris.coords.DimCoord`.
        * grid_x_coord:
            The :class:`iris.coords.DimCoord` for the new grid's X
            coordinate.
        * grid_y_coord:
            The :class:`iris.coords.DimCoord` for the new grid's Y
            coordinate.
        * sample_grid_x:
            A 2-dimensional array of sample X values.
        * sample_grid_y:
            A 2-dimensional array of sample Y values.
        * regrid_callback:
            The routine that will be used to calculate the interpolated
            values of any reference surfaces.

        Returns:
            The new, regridded Cube.

        """
        #
        # XXX: At the moment requires to be a static method as used by
        # experimental regrid_area_weighted_rectilinear_src_and_grid
        #
        # Create a result cube with the appropriate metadata
        result = iris.cube.Cube(data)
        result.metadata = copy.deepcopy(src.metadata)

        # Copy across all the coordinates which don't span the grid.
        # Record a mapping from old coordinate IDs to new coordinates,
        # for subsequent use in creating updated aux_factories.
        coord_mapping = {}

        def copy_coords(src_coords, add_method):
            for coord in src_coords:
                dims = src.coord_dims(coord)
                if coord == src_x_coord:
                    coord = grid_x_coord
                elif coord == src_y_coord:
                    coord = grid_y_coord
                elif x_dim in dims or y_dim in dims:
                    continue
                result_coord = coord.copy()
                add_method(result_coord, dims)
                coord_mapping[id(coord)] = result_coord

        copy_coords(src.dim_coords, result.add_dim_coord)
        copy_coords(src.aux_coords, result.add_aux_coord)

        def regrid_reference_surface(
            src_surface_coord,
            surface_dims,
            x_dim,
            y_dim,
            src_x_coord,
            src_y_coord,
            sample_grid_x,
            sample_grid_y,
            regrid_callback,
        ):
            # Determine which of the reference surface's dimensions span the X
            # and Y dimensions of the source cube.
            surface_x_dim = surface_dims.index(x_dim)
            surface_y_dim = surface_dims.index(y_dim)
            surface = regrid_callback(
                src_surface_coord.points,
                surface_x_dim,
                surface_y_dim,
                src_x_coord,
                src_y_coord,
                sample_grid_x,
                sample_grid_y,
            )
            surface_coord = src_surface_coord.copy(surface)
            return surface_coord

        # Copy across any AuxFactory instances, and regrid their reference
        # surfaces where required.
        for factory in src.aux_factories:
            for coord in factory.dependencies.values():
                if coord is None:
                    continue
                dims = src.coord_dims(coord)
                if x_dim in dims and y_dim in dims:
                    result_coord = regrid_reference_surface(
                        coord,
                        dims,
                        x_dim,
                        y_dim,
                        src_x_coord,
                        src_y_coord,
                        sample_grid_x,
                        sample_grid_y,
                        regrid_callback,
                    )
                    result.add_aux_coord(result_coord, dims)
                    coord_mapping[id(coord)] = result_coord
            try:
                result.add_aux_factory(factory.updated(coord_mapping))
            except KeyError:
                msg = (
                    "Cannot update aux_factory {!r} because of dropped"
                    " coordinates.".format(factory.name())
                )
                warnings.warn(msg)
        return result

    def _check_units(self, coord):
        if coord.coord_system is None:
            # No restriction on units.
            pass
        elif isinstance(
            coord.coord_system,
            (iris.coord_systems.GeogCS, iris.coord_systems.RotatedGeogCS),
        ):
            # Units for lat-lon or rotated pole must be 'degrees'. Note
            # that 'degrees_east' etc. are equal to 'degrees'.
            if coord.units != "degrees":
                msg = (
                    "Unsupported units for coordinate system. "
                    "Expected 'degrees' got {!r}.".format(coord.units)
                )
                raise ValueError(msg)
        else:
            # Units for other coord systems must be equal to metres.
            if coord.units != "m":
                msg = (
                    "Unsupported units for coordinate system. "
                    "Expected 'metres' got {!r}.".format(coord.units)
                )
                raise ValueError(msg)

    def __call__(self, src):
        """
        Regrid this :class:`~iris.cube.Cube` on to the target grid of
        this :class:`RectilinearRegridder`.

        The given cube must be defined with the same grid as the source
        grid used to create this :class:`RectilinearRegridder`.

        If the source cube has lazy data, the returned cube will also
        have lazy data.

        Args:

        * src:
            A :class:`~iris.cube.Cube` to be regridded.

        Returns:
            A cube defined with the horizontal dimensions of the target
            and the other dimensions from this cube. The data values of
            this cube will be converted to values on the new grid using
            either nearest-neighbour or linear interpolation.

        .. note::

            If the source cube has lazy data,
            `chunks <https://docs.dask.org/en/latest/array-chunks.html>`__
            in the horizontal dimensions will be combined before regridding.

        """
        # Validity checks.
        if not isinstance(src, iris.cube.Cube):
            raise TypeError("'src' must be a Cube")
        if get_xy_dim_coords(src) != self._src_grid:
            raise ValueError(
                "The given cube is not defined on the same "
                "source grid as this regridder."
            )

        src_x_coord, src_y_coord = get_xy_dim_coords(src)
        grid_x_coord, grid_y_coord = self._tgt_grid
        src_cs = src_x_coord.coord_system
        grid_cs = grid_x_coord.coord_system

        if src_cs is None and grid_cs is None:
            if not (
                src_x_coord.is_compatible(grid_x_coord)
                and src_y_coord.is_compatible(grid_y_coord)
            ):
                raise ValueError(
                    "The rectilinear grid coordinates of the "
                    "given cube and target grid have no "
                    "coordinate system but they do not have "
                    "matching coordinate metadata."
                )
        elif src_cs is None or grid_cs is None:
            raise ValueError(
                "The rectilinear grid coordinates of the given "
                "cube and target grid must either both have "
                "coordinate systems or both have no coordinate "
                "system but with matching coordinate metadata."
            )

        # Check the source grid units.
        for coord in (src_x_coord, src_y_coord):
            self._check_units(coord)

        # Convert the grid to a 2D sample grid in the src CRS.
        sample_grid = self._sample_grid(src_cs, grid_x_coord, grid_y_coord)
        sample_grid_x, sample_grid_y = sample_grid

        # Compute the interpolated data values.
        x_dim = src.coord_dims(src_x_coord)[0]
        y_dim = src.coord_dims(src_y_coord)[0]

        # Define regrid function
        regrid = functools.partial(
            self._regrid,
            x_dim=x_dim,
            y_dim=y_dim,
            src_x_coord=src_x_coord,
            src_y_coord=src_y_coord,
            sample_grid_x=sample_grid_x,
            sample_grid_y=sample_grid_y,
            method=self._method,
            extrapolation_mode=self._extrapolation_mode,
        )

        data = map_complete_blocks(
            src, regrid, (y_dim, x_dim), sample_grid_x.shape
        )

        # Wrap up the data as a Cube.
        regrid_callback = functools.partial(
            self._regrid, method=self._method, extrapolation_mode="nan"
        )
        result = self._create_cube(
            data,
            src,
            x_dim,
            y_dim,
            src_x_coord,
            src_y_coord,
            grid_x_coord,
            grid_y_coord,
            sample_grid_x,
            sample_grid_y,
            regrid_callback,
        )
        return result



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Support for "fast" loading of structured UM files in iris load functions,
i.e. :meth:`iris.load` and its associates.

This provides a context manager to enable structured loading via all the iris
load function variants.  It is made public in :mod:`iris.fileformats.um`.

This works with either PP or Fieldsfiles.
The various existing PP and FF format loaders all call into
:meth:`iris.fileformats.pp._load_cubes_variable_loader`.
When enabled, that function calls private functions in this module, to replace
the 'normal' calls with 'structured' loading.

At present, there is *no* public low-level fields-to-cube interface, equivalent
to the pp "as_pairs" functions.

"""

from contextlib import contextmanager
import threading
import os.path

import numpy as np

# Be minimal about what we import from iris, to avoid circular imports.
# Below, other parts of iris.fileformats are accessed via deferred imports.
from iris.coords import DimCoord
from iris.cube import CubeList
from iris.exceptions import TranslationError
from iris.fileformats.um._fast_load_structured_fields import (
    BasicFieldCollation,
    group_structured_fields,
)

# Strings to identify the PP and FF file format handler specs.
_FF_SPEC_NAME = "UM Fieldsfile"
_PP_SPEC_NAME = "UM Post Processing file"


class FieldCollation(BasicFieldCollation):
    # This class specialises the BasicFieldCollation by adding the file-index
    # and file-path concepts.
    # This preserves the more abstract scope of the original 'FieldCollation'
    # class, now renamed 'BasicFieldCollation'.

    def __init__(self, fields, filepath):
        """
        Args:

        * fields (iterable of :class:`iris.fileformats.pp.PPField`):
            The fields in the collation.

        * filepath (string):
            The path of the file the collation is loaded from.

        """
        super().__init__(fields)
        self._load_filepath = filepath

    @property
    def data_filepath(self):
        return self._load_filepath

    @property
    def data_field_indices(self):
        """
        Field indices of the contained PPFields in the input file.

        This records the original file location of the individual data fields
        contained, within the input datafile.

        Returns:
            An integer array of shape `self.vector_dims_shape`.

        """
        # Get shape :  N.B. this calculates (and caches) the structure.
        vector_dims_shape = self.vector_dims_shape
        # Get index-in-file of each contained field.
        indices = np.array(
            [field._index_in_structured_load_file for field in self._fields],
            dtype=np.int64,
        )
        return indices.reshape(vector_dims_shape)


def _basic_load_function(filename, pp_filter=None, **kwargs):
    # The low-level 'fields from filename' loader.
    #
    # This is the 'loader.generator' in the control structure passed to the
    # generic rules code, :meth:`iris.fileformats.rules.load_cubes`.
    #
    # As called by generic rules code, this replaces pp.load (and the like).
    # It yields a sequence of "fields", but in this case the 'fields' are
    # :class:`iris.fileformats.um._fast_load_structured_fields.FieldCollation`
    # objects.
    #
    # NOTE: so, a FieldCollation is passed as the 'field' in user callbacks.
    #
    # Also in our case, we need to support the basic single-field filtering
    # operation that speeds up phenomenon selection.
    # Therefore, the actual loader will pass this as the 'pp_filter' keyword,
    # when it is present.
    # Additional load keywords are 'passed on' to the lower-level function.

    # Helper function to select the correct fields loader call.
    def _select_raw_fields_loader(fname):
        # Return the PPfield loading function for a file name.
        #
        # This decides whether the underlying file is an FF or PP file.
        # Because it would be too awkward to modify the whole iris loading
        # callchain to "pass down" the file format, this function instead
        # 'recreates' that information by calling the format picker again.
        # NOTE: this may be inefficient, especially for web resources.
        from iris.fileformats import FORMAT_AGENT
        from iris.fileformats.pp import load as pp_load
        from iris.fileformats.um import um_to_pp

        with open(fname, "rb") as fh:
            spec = FORMAT_AGENT.get_spec(os.path.basename(fname), fh)
        if spec.name.startswith(_FF_SPEC_NAME):
            loader = um_to_pp
        elif spec.name.startswith(_PP_SPEC_NAME):
            loader = pp_load
        else:
            emsg = "Require {!r} to be a structured FieldsFile or a PP file."
            raise ValueError(emsg.format(fname))
        return loader

    loader = _select_raw_fields_loader(filename)

    def iter_fields_decorated_with_load_indices(fields_iter):
        for i_field, field in enumerate(fields_iter):
            field._index_in_structured_load_file = i_field
            yield field

    fields = iter_fields_decorated_with_load_indices(
        field
        for field in loader(filename, **kwargs)
        if pp_filter is None or pp_filter(field)
    )

    return group_structured_fields(
        fields, collation_class=FieldCollation, filepath=filename
    )


# Define the preferred order of candidate dimension coordinates, as used by
# _convert_collation.
_HINT_COORDS = ["time", "forecast_reference_time", "model_level_number"]
_HINTS = {name: i for i, name in zip(range(len(_HINT_COORDS)), _HINT_COORDS)}


def _convert_collation(collation):
    """
    Converts a FieldCollation into the corresponding items of Cube
    metadata.

    Args:

    * collation:
        A FieldCollation object.

    Returns:
        A :class:`iris.fileformats.rules.ConversionMetadata` object.

    .. note:
        This is the 'loader.converter', in the control structure passed to the
        generic rules code, :meth:`iris.fileformats.rules.load_cubes`.

    """
    from iris.fileformats.rules import ConversionMetadata
    from iris.fileformats.pp_load_rules import (
        _convert_time_coords,
        _convert_vertical_coords,
        _convert_scalar_realization_coords,
        _convert_scalar_pseudo_level_coords,
        _all_other_rules,
    )

    # For all the scalar conversions, all fields in the collation will
    # give the same result, so the choice is arbitrary.
    field = collation.fields[0]

    # Call "all other" rules.
    (
        references,
        standard_name,
        long_name,
        units,
        attributes,
        cell_methods,
        dim_coords_and_dims,
        aux_coords_and_dims,
    ) = _all_other_rules(field)

    # Adjust any dimension bindings to account for the extra leading
    # dimensions added by the collation.
    if collation.vector_dims_shape:

        def _adjust_dims(coords_and_dims, n_dims):
            def adjust(dims):
                if dims is not None:
                    dims += n_dims
                return dims

            return [(coord, adjust(dims)) for coord, dims in coords_and_dims]

        n_collation_dims = len(collation.vector_dims_shape)
        dim_coords_and_dims = _adjust_dims(
            dim_coords_and_dims, n_collation_dims
        )
        aux_coords_and_dims = _adjust_dims(
            aux_coords_and_dims, n_collation_dims
        )

    # Dimensions to which we've already assigned dimension coordinates.
    dim_coord_dims = set()

    # Helper call to choose which coords are dimensions and which auxiliary.
    def _bind_coords(
        coords_and_dims,
        dim_coord_dims,
        dim_coords_and_dims,
        aux_coords_and_dims,
    ):
        def key_func(item):
            return _HINTS.get(item[0].name(), len(_HINTS))

        # Target the first DimCoord for a dimension at dim_coords,
        # and target everything else at aux_coords.
        for coord, dims in sorted(coords_and_dims, key=key_func):
            if (
                isinstance(coord, DimCoord)
                and dims is not None
                and len(dims) == 1
                and dims[0] not in dim_coord_dims
            ):
                dim_coords_and_dims.append((coord, dims))
                dim_coord_dims.add(dims[0])
            else:
                aux_coords_and_dims.append((coord, dims))

    # Call "time" rules.
    #
    # For "normal" (non-cross-sectional) time values.
    vector_headers = collation.element_arrays_and_dims
    # If the collation doesn't define a vector of values for a
    # particular header then it must be constant over all fields in the
    # collation. In which case it's safe to get the value from any field.
    t1, t1_dims = vector_headers.get("t1", (field.t1, ()))
    t2, t2_dims = vector_headers.get("t2", (field.t2, ()))
    lbft, lbft_dims = vector_headers.get("lbft", (field.lbft, ()))
    coords_and_dims = _convert_time_coords(
        field.lbcode,
        field.lbtim,
        field.time_unit("hours"),
        t1,
        t2,
        lbft,
        t1_dims,
        t2_dims,
        lbft_dims,
    )
    # Bind resulting coordinates to dimensions, where suitable.
    _bind_coords(
        coords_and_dims,
        dim_coord_dims,
        dim_coords_and_dims,
        aux_coords_and_dims,
    )

    # Call "vertical" rules.
    #
    # "Normal" (non-cross-sectional) vertical levels
    blev, blev_dims = vector_headers.get("blev", (field.blev, ()))
    lblev, lblev_dims = vector_headers.get("lblev", (field.lblev, ()))
    bhlev, bhlev_dims = vector_headers.get("bhlev", (field.bhlev, ()))
    bhrlev, bhrlev_dims = vector_headers.get("bhrlev", (field.bhrlev, ()))
    brsvd1, brsvd1_dims = vector_headers.get("brsvd1", (field.brsvd[0], ()))
    brsvd2, brsvd2_dims = vector_headers.get("brsvd2", (field.brsvd[1], ()))
    brlev, brlev_dims = vector_headers.get("brlev", (field.brlev, ()))
    # Find all the non-trivial dimension values
    dims = set(
        filter(
            None,
            [
                blev_dims,
                lblev_dims,
                bhlev_dims,
                bhrlev_dims,
                brsvd1_dims,
                brsvd2_dims,
                brlev_dims,
            ],
        )
    )
    if len(dims) > 1:
        raise TranslationError(
            "Unsupported multiple values for vertical " "dimension."
        )
    if dims:
        v_dims = dims.pop()
        if len(v_dims) > 1:
            raise TranslationError(
                "Unsupported multi-dimension vertical " "headers."
            )
    else:
        v_dims = ()
    coords_and_dims, factories = _convert_vertical_coords(
        field.lbcode,
        field.lbvc,
        blev,
        lblev,
        field.stash,
        bhlev,
        bhrlev,
        brsvd1,
        brsvd2,
        brlev,
        v_dims,
    )
    # Bind resulting coordinates to dimensions, where suitable.
    _bind_coords(
        coords_and_dims,
        dim_coord_dims,
        dim_coords_and_dims,
        aux_coords_and_dims,
    )

    # Realization (aka ensemble) (--> scalar coordinates)
    aux_coords_and_dims.extend(
        _convert_scalar_realization_coords(lbrsvd4=field.lbrsvd[3])
    )

    # Pseudo-level coordinate (--> scalar coordinates)
    aux_coords_and_dims.extend(
        _convert_scalar_pseudo_level_coords(lbuser5=field.lbuser[4])
    )

    return ConversionMetadata(
        factories,
        references,
        standard_name,
        long_name,
        units,
        attributes,
        cell_methods,
        dim_coords_and_dims,
        aux_coords_and_dims,
    )


def _combine_structured_cubes(cubes):
    # Combine structured cubes from different sourcefiles, in the style of
    # merge/concatenate.
    #
    # Because standard Cube.merge employed in loading can't do this.
    if STRUCTURED_LOAD_CONTROLS.structured_load_is_raw:
        # Cross-file concatenate is disabled, during a "load_raw" call.
        result = cubes
    else:
        result = iter(CubeList(cubes).concatenate())
    return result


class StructuredLoadFlags(threading.local):
    # A thread-safe object to control structured loading operations.
    # The object properties are the control flags.
    #
    # Inheriting from 'threading.local' provides a *separate* set of the
    # object properties for each thread.
    def __init__(self):
        # Control whether iris load functions do structured loads.
        self.loads_use_structured = False
        # Control whether structured load 'combine' is enabled.
        self.structured_load_is_raw = False

    @contextmanager
    def context(self, loads_use_structured=None, structured_load_is_raw=None):
        # Snapshot current states, for restoration afterwards.
        old_structured = self.loads_use_structured
        old_raw_load = self.structured_load_is_raw
        try:
            # Set flags for duration, as requested.
            if loads_use_structured is not None:
                self.loads_use_structured = loads_use_structured
            if structured_load_is_raw is not None:
                self.structured_load_is_raw = structured_load_is_raw
            # Yield to caller operation.
            yield
        finally:
            # Restore entry state of flags.
            self.loads_use_structured = old_structured
            self.structured_load_is_raw = old_raw_load


# A singleton structured-load-control object.
# Used in :meth:`iris.fileformats.pp._load_cubes_variable_loader`.
STRUCTURED_LOAD_CONTROLS = StructuredLoadFlags()


@contextmanager
def structured_um_loading():
    """
    Load cubes from structured UM Fieldsfile and PP files.

    "Structured" loading is a streamlined, fast load operation, to be used
    **only** on fieldsfiles or PP files whose fields repeat regularly over
    the same vertical levels and times (see full details below).

    This method is a context manager which enables an alternative loading
    mechanism for 'structured' UM files, providing much faster load times.
    Within the scope of the context manager, this affects all standard Iris
    load functions (:func:`~iris.load`, :func:`~iris.load_cube`,
    :func:`~iris.load_cubes` and :func:`~iris.load_raw`), when loading from UM
    format files (PP or fieldsfiles).

    For example:

        >>> import iris
        >>> filepath = iris.sample_data_path('uk_hires.pp')
        >>> from iris.fileformats.um import structured_um_loading
        >>> with structured_um_loading():
        ...     cube = iris.load_cube(filepath, 'air_potential_temperature')
        ...
        >>> cube
        <iris 'Cube' of air_potential_temperature / (K) \
(time: 3; model_level_number: 7; grid_latitude: 204; grid_longitude: 187)>

    The results from this are normally equivalent to those generated by
    :func:`iris.load`, but the operation is substantially faster for input
    which is structured.

    For calls other than :meth:`~iris.load_raw`, the resulting cubes are
    concatenated over all the input files, so there is normally just one
    output cube per phenomenon.

    However, actual loaded results are somewhat different from non-structured
    loads in many cases, and in a variety of ways.  Most commonly, dimension
    ordering and the choice of dimension coordinates are often different.

    Use of load callbacks:

        When a user callback function is used with structured-loading, it is
        called in a somewhat different way than in a 'normal' load :
        The callback is called once for each basic *structured* cube loaded,
        which is normally the whole of one phenomenon from a single input file.
        In particular, the callback's "field" argument is a
        :class:`~iris.fileformats.um.FieldCollation`, from which "field.fields"
        gives a *list* of PPFields from which that cube was built, and the
        properties "field.load_filepath" and "field.load_file_indices"
        reference the original file locations of the cube data.
        The code required is therefore different from a 'normal' callback.
        For an example of this, see `this example in the Iris test code
        <https://github.com/SciTools/iris/
        blob/ddb46f78e54b6ef4110357dfe9cfcffa7d186d90/
        lib/iris/tests/integration/fast_load/test_fast_load.py#L409>`_.

    Notes on applicability:

        For results to be **correct and reliable**, the input files must
        conform to the following requirements :

        *  the file must contain fields for all possible combinations of the
           vertical levels and time points found in the file.

        *  the fields must occur in a regular repeating order within the file,
           within the fields of each phenomenon.

           For example: a sequence of fields for NV vertical levels, repeated
           for NP different forecast periods, repeated for NT different
           forecast times.

        *  all other metadata must be identical across all fields of the same
           phenomenon.

        Each group of fields with the same values of LBUSER4, LBUSER7 and
        LBPROC is identified as a separate phenomenon:  These groups are
        processed independently and returned as separate result cubes.
        The need for a regular sequence of fields applies separately to the
        fields of each phenomenon, such that different phenomena may have
        different field structures, and can be interleaved in any way at all.

        .. note::

             At present, fields with different values of 'LBUSER5'
             (pseudo-level) are *also* treated internally as different
             phenomena, yielding a raw cube per level.
             The effects of this are not normally noticed, as the resulting
             multiple raw cubes merge together again in a 'normal' load.
             However, it is not an ideal solution as operation is less
             efficient (in particular, slower) :  it is done to avoid a
             limitation in the underlying code which would otherwise load data
             on pseudo-levels incorrectly.  In future, this may be corrected.

    Known current shortcomings:

        * orography fields may be returned with extra dimensions, e.g. time,
          where multiple fields exist in an input file.

        * if some input files contain a *single* coordinate value while others
          contain *multiple* values, these will not be merged into a single
          cube over all input files :  Instead, the single- and multiple-valued
          sets will typically produce two separate cubes with overlapping
          coordinates.

          * this can be worked around by loading files individually, or with
            :meth:`~iris.load_raw`, and merging/concatenating explicitly.

    .. note::

        The resulting time-related coordinates ('time', 'forecast_time' and
        'forecast_period') may be mapped to shared cube dimensions and in some
        cases can also be multidimensional.  However, the vertical level
        information *must* have a simple one-dimensional structure, independent
        of the time points, otherwise an error will be raised.

    .. note::

        Where input data does *not* have a fully regular arrangement, the
        corresponding result cube will have a single anonymous extra dimension
        which indexes over all the input fields.

        This can happen if, for example, some fields are missing; or have
        slightly different metadata; or appear out of order in the file.

    .. warning::

        Restrictions and limitations:

        Any non-regular metadata variation in the input should be strictly
        avoided, as not all irregularities are detected, which can cause
        erroneous results.

        Various field header words which can in some cases vary are assumed to
        have a constant value throughout a given phenomenon.  This is **not**
        checked, and can lead to erroneous results if it is not the case.
        Header elements of potential concern include LBTIM, LBCODE, LBVC and
        LBRSVD4 (ensemble number).

    """
    with STRUCTURED_LOAD_CONTROLS.context(loads_use_structured=True):
        yield


@contextmanager
def _raw_structured_loading():
    """
    Private context manager called by :func:`iris.load_raw` to prevent
    structured loading from concatenating its result cubes in that case.

    """
    with STRUCTURED_LOAD_CONTROLS.context(structured_load_is_raw=True):
        yield



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Various utilities related to geometric operations.

.. note::
    This module requires :mod:`shapely`.

"""

import warnings

from shapely.geometry import Polygon

import numpy as np

import iris.exceptions


def _extract_relevant_cube_slice(cube, geometry):
    """
    Given a shapely geometry object, this helper method returns
    the tuple
    (subcube, x_coord_of_subcube, y_coord_of_subcube,
     (min_x_index, min_y_index, max_x_index, max_y_index))

    If cube and geometry don't overlap, returns None.

    """

    # Validate the input parameters
    if not cube.coords(axis="x") or not cube.coords(axis="y"):
        raise ValueError("The cube must contain x and y axes.")

    x_coords = cube.coords(axis="x")
    y_coords = cube.coords(axis="y")
    if len(x_coords) != 1 or len(y_coords) != 1:
        raise ValueError(
            "The cube must contain one, and only one, coordinate "
            "for each of the x and y axes."
        )

    x_coord = x_coords[0]
    y_coord = y_coords[0]
    if not (x_coord.has_bounds() and y_coord.has_bounds()):
        raise ValueError("Both horizontal coordinates must have bounds.")

    if x_coord.ndim != 1:
        raise iris.exceptions.CoordinateMultiDimError(x_coord)
    if y_coord.ndim != 1:
        raise iris.exceptions.CoordinateMultiDimError(y_coord)

    # bounds of cube dimensions
    x_bounds = x_coord.bounds
    y_bounds = y_coord.bounds

    # identify ascending/descending coordinate dimensions
    x_ascending = x_coord.points[1] - x_coord.points[0] > 0.0
    y_ascending = y_coord.points[1] - y_coord.points[0] > 0.0

    # identify upper/lower bounds of coordinate dimensions
    x_bounds_lower = x_bounds[:, 0] if x_ascending else x_bounds[:, 1]
    y_bounds_lower = y_bounds[:, 0] if y_ascending else y_bounds[:, 1]
    x_bounds_upper = x_bounds[:, 1] if x_ascending else x_bounds[:, 0]
    y_bounds_upper = y_bounds[:, 1] if y_ascending else y_bounds[:, 0]

    # find indices of coordinate bounds to fully cover geometry
    x_min_geom, y_min_geom, x_max_geom, y_max_geom = geometry.bounds
    try:
        x_min_ix = np.where(x_bounds_lower <= x_min_geom)[0]
        x_min_ix = x_min_ix[np.argmax(x_bounds_lower[x_min_ix])]
    except ValueError:
        warnings.warn(
            "The geometry exceeds the cube's x dimension at the " "lower end.",
            UserWarning,
        )
        x_min_ix = 0 if x_ascending else x_coord.points.size - 1

    try:
        x_max_ix = np.where(x_bounds_upper >= x_max_geom)[0]
        x_max_ix = x_max_ix[np.argmin(x_bounds_upper[x_max_ix])]
    except ValueError:
        warnings.warn(
            "The geometry exceeds the cube's x dimension at the " "upper end.",
            UserWarning,
        )
        x_max_ix = x_coord.points.size - 1 if x_ascending else 0

    try:
        y_min_ix = np.where(y_bounds_lower <= y_min_geom)[0]
        y_min_ix = y_min_ix[np.argmax(y_bounds_lower[y_min_ix])]
    except ValueError:
        warnings.warn(
            "The geometry exceeds the cube's y dimension at the " "lower end.",
            UserWarning,
        )
        y_min_ix = 0 if y_ascending else y_coord.points.size - 1

    try:
        y_max_ix = np.where(y_bounds_upper >= y_max_geom)[0]
        y_max_ix = y_max_ix[np.argmin(y_bounds_upper[y_max_ix])]
    except ValueError:
        warnings.warn(
            "The geometry exceeds the cube's y dimension at the " "upper end.",
            UserWarning,
        )
        y_max_ix = y_coord.points.size - 1 if y_ascending else 0

    # extract coordinate values at these indices
    x_min = x_bounds_lower[x_min_ix]
    x_max = x_bounds_upper[x_max_ix]
    y_min = y_bounds_lower[y_min_ix]
    y_max = y_bounds_upper[y_max_ix]

    # switch min and max if necessary, to create slice objects later on
    if x_min_ix > x_max_ix:
        x_min_ix, x_max_ix = x_max_ix, x_min_ix
    if y_min_ix > y_max_ix:
        y_min_ix, y_max_ix = y_max_ix, y_min_ix
    bnds_ix = x_min_ix, y_min_ix, x_max_ix, y_max_ix

    # cut the relevant part from the original cube
    coord_constr = {
        x_coord.name(): lambda x: x_min <= x.point <= x_max,
        y_coord.name(): lambda y: y_min <= y.point <= y_max,
    }
    constraint = iris.Constraint(coord_values=coord_constr)
    subcube = cube.extract(constraint)

    if subcube is None:
        return None
    else:
        x_coord = subcube.coord(axis="x")
        y_coord = subcube.coord(axis="y")
        return subcube, x_coord, y_coord, bnds_ix


def geometry_area_weights(cube, geometry, normalize=False):
    """
    Returns the array of weights corresponding to the area of overlap between
    the cells of cube's horizontal grid, and the given shapely geometry.

    The returned array is suitable for use with :const:`iris.analysis.MEAN`.

    The cube must have bounded horizontal coordinates.

    .. note::
        This routine works in Euclidean space. Area calculations do not
        account for the curvature of the Earth. And care must be taken to
        ensure any longitude values are expressed over a suitable interval.

    .. note::
        This routine currently does not handle all out-of-bounds cases
        correctly. In cases where both the coordinate bounds and the
        geometry's bounds lie outside the physically realistic range
        (i.e., abs(latitude) > 90., as it is commonly the case when
        bounds are constructed via guess_bounds()), the weights
        calculation might be wrong. In this case, a UserWarning will
        be issued.

    Args:

    * cube (:class:`iris.cube.Cube`):
        A Cube containing a bounded, horizontal grid definition.
    * geometry (a shapely geometry instance):
        The geometry of interest. To produce meaningful results this geometry
        must have a non-zero area. Typically a Polygon or MultiPolygon.

    Kwargs:

    * normalize:
        Calculate each individual cell weight as the cell area overlap between
        the cell and the given shapely geometry divided by the total cell area.
        Default is False.

    """

    # extract smallest subcube containing geometry
    shape = cube.shape
    extraction_results = _extract_relevant_cube_slice(cube, geometry)

    # test if there is overlap between cube and geometry
    if extraction_results is None:
        return np.zeros(shape)

    subcube, subx_coord, suby_coord, bnds_ix = extraction_results
    x_min_ix, y_min_ix, x_max_ix, y_max_ix = bnds_ix

    # prepare the weights array
    subshape = list(cube.shape)
    x_dim = cube.coord_dims(subx_coord)[0]
    y_dim = cube.coord_dims(suby_coord)[0]
    subshape[x_dim] = subx_coord.shape[0]
    subshape[y_dim] = suby_coord.shape[0]
    subx_bounds = subx_coord.bounds
    suby_bounds = suby_coord.bounds
    subweights = np.empty(subshape, np.float32)

    # calculate the area weights
    for nd_index in np.ndindex(subweights.shape):
        xi = nd_index[x_dim]
        yi = nd_index[y_dim]
        x0, x1 = subx_bounds[xi]
        y0, y1 = suby_bounds[yi]
        polygon = Polygon([(x0, y0), (x0, y1), (x1, y1), (x1, y0)])
        subweights[nd_index] = polygon.intersection(geometry).area
        if normalize:
            subweights[nd_index] /= polygon.area

    # pad the calculated weights with zeros to match original cube shape
    weights = np.zeros(shape, np.float32)
    slices = []
    for i in range(weights.ndim):
        if i == x_dim:
            slices.append(slice(x_min_ix, x_max_ix + 1))
        elif i == y_dim:
            slices.append(slice(y_min_ix, y_max_ix + 1))
        else:
            slices.append(slice(None))

    weights[tuple(slices)] = subweights

    # Fix for the limitation of iris.analysis.MEAN weights handling.
    # Broadcast the array to the full shape of the cube
    weights = np.broadcast_arrays(weights, cube.data)[0]

    return weights



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.reorder_bounds_data`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    reorder_bounds_data


class Test(tests.IrisTest):
    def test_fastest_varying(self):
        bounds_data = np.arange(24).reshape(2, 3, 4)
        cf_bounds_var = mock.Mock(dimensions=('foo', 'bar', 'nv'),
                                  cf_name='wibble_bnds')
        cf_coord_var = mock.Mock(dimensions=('foo', 'bar'))

        res = reorder_bounds_data(bounds_data, cf_bounds_var, cf_coord_var)
        # Vertex dimension (nv) is already at the end.
        self.assertArrayEqual(res, bounds_data)

    def test_slowest_varying(self):
        bounds_data = np.arange(24).reshape(4, 2, 3)
        cf_bounds_var = mock.Mock(dimensions=('nv', 'foo', 'bar'))
        cf_coord_var = mock.Mock(dimensions=('foo', 'bar'))

        res = reorder_bounds_data(bounds_data, cf_bounds_var, cf_coord_var)
        # Move zeroth dimension (nv) to the end.
        expected = np.rollaxis(bounds_data, 0, bounds_data.ndim)
        self.assertArrayEqual(res, expected)

    def test_different_dim_names(self):
        bounds_data = np.arange(24).reshape(2, 3, 4)
        cf_bounds_var = mock.Mock(dimensions=('foo', 'bar', 'nv'),
                                  cf_name='wibble_bnds')
        cf_coord_var = mock.Mock(dimensions=('x', 'y'), cf_name='wibble')
        with self.assertRaisesRegex(ValueError, 'dimension names'):
            reorder_bounds_data(bounds_data, cf_bounds_var, cf_coord_var)


if __name__ == '__main__':
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :func:`iris.common.mixin._get_valid_standard_name`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.common.mixin import _get_valid_standard_name


class Test(tests.IrisTest):
    def setUp(self):
        self.emsg = "'{}' is not a valid standard_name"

    def test_pass_thru_none(self):
        name = None
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_pass_thru_empty(self):
        name = ""
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_pass_thru_whitespace(self):
        name = "       "
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_valid_standard_name(self):
        name = "air_temperature"
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_standard_name_alias(self):
        name = "atmosphere_optical_thickness_due_to_pm1_ambient_aerosol"
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_invalid_standard_name(self):
        name = "not_a_standard_name"
        with self.assertRaisesRegex(ValueError, self.emsg.format(name)):
            _get_valid_standard_name(name)

    def test_valid_standard_name_valid_modifier(self):
        name = "air_temperature standard_error"
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_valid_standard_name_valid_modifier_extra_spaces(self):
        name = "air_temperature      standard_error"
        self.assertEqual(_get_valid_standard_name(name), name)

    def test_invalid_standard_name_valid_modifier(self):
        name = "not_a_standard_name standard_error"
        with self.assertRaisesRegex(ValueError, self.emsg.format(name)):
            _get_valid_standard_name(name)

    def test_valid_standard_invalid_name_modifier(self):
        name = "air_temperature extra_names standard_error"
        with self.assertRaisesRegex(ValueError, self.emsg.format(name)):
            _get_valid_standard_name(name)

    def test_valid_standard_valid_name_modifier_extra_names(self):
        name = "air_temperature standard_error extra words"
        with self.assertRaisesRegex(ValueError, self.emsg.format(name)):
            _get_valid_standard_name(name)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the constrained cube loading mechanism.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import datetime

import iris
from iris import AttributeConstraint, NameConstraint
import iris.tests.stock as stock


SN_AIR_POTENTIAL_TEMPERATURE = "air_potential_temperature"
SN_SPECIFIC_HUMIDITY = "specific_humidity"


# TODO: Workaround, pending #1262
def workaround_pending_1262(cubes):
    """Reverse the cube if sigma was chosen as a dim_coord."""
    for i, cube in enumerate(cubes):
        ml = cube.coord("model_level_number").points
        if ml[0] > ml[1]:
            cubes[i] = cube[::-1]


@tests.skip_data
class TestSimple(tests.IrisTest):
    def setUp(self):
        names = ["grid_latitude", "grid_longitude"]
        self.slices = iris.cube.CubeList(stock.realistic_4d().slices(names))

    def test_constraints(self):
        constraint = iris.Constraint(model_level_number=10)
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 6)

        constraint = iris.Constraint(model_level_number=[10, 22])
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 2 * 6)

        constraint = iris.Constraint(
            model_level_number=lambda c: (c > 30) | (c <= 3)
        )
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 43 * 6)

        constraint = iris.Constraint(
            coord_values={"model_level_number": lambda c: c > 1000}
        )
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 0)

        constraint = iris.Constraint(model_level_number=10) & iris.Constraint(
            time=datetime.datetime(2009, 9, 9, 18, 0)
        )
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 1)

        constraint = iris.Constraint(SN_AIR_POTENTIAL_TEMPERATURE)
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 70 * 6)

    def test_mismatched_type(self):
        constraint = iris.Constraint(model_level_number="aardvark")
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 0)

    def test_cell(self):
        cell = iris.coords.Cell(10)
        constraint = iris.Constraint(model_level_number=cell)
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 6)

    def test_cell_equal_bounds(self):
        cell = self.slices[0].coord("level_height").cell(0)
        constraint = iris.Constraint(level_height=cell)
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 6)

    def test_cell_different_bounds(self):
        cell = iris.coords.Cell(10, bound=(9, 11))
        constraint = iris.Constraint(model_level_number=cell)
        sub_list = self.slices.extract(constraint)
        self.assertEqual(len(sub_list), 0)


class TestMixin:
    """
    Mix-in class for attributes & utilities common to the "normal" and "strict" test cases.

    """

    def setUp(self):
        self.dec_path = tests.get_data_path(
            ["PP", "globClim1", "dec_subset.pp"]
        )
        self.theta_path = tests.get_data_path(["PP", "globClim1", "theta.pp"])

        self.humidity = iris.Constraint(SN_SPECIFIC_HUMIDITY)
        self.theta = iris.Constraint(SN_AIR_POTENTIAL_TEMPERATURE)

        # Coord based constraints
        self.level_10 = iris.Constraint(model_level_number=10)
        self.level_22 = iris.Constraint(model_level_number=22)

        # Value based coord constraint
        self.level_30 = iris.Constraint(model_level_number=30)
        self.level_gt_30_le_3 = iris.Constraint(
            model_level_number=lambda c: (c > 30) | (c <= 3)
        )
        self.invalid_inequality = iris.Constraint(
            coord_values={"model_level_number": lambda c: c > 1000}
        )

        # bound based coord constraint
        self.level_height_of_model_level_number_10 = iris.Constraint(
            level_height=1900
        )
        self.model_level_number_10_22 = iris.Constraint(
            model_level_number=[10, 22]
        )

        # Invalid constraints
        self.pressure_950 = iris.Constraint(model_level_number=950)

        self.lat_30 = iris.Constraint(latitude=30)
        self.lat_gt_45 = iris.Constraint(latitude=lambda c: c > 45)


class RelaxedConstraintMixin(TestMixin):
    @staticmethod
    def fixup_sigma_to_be_aux(cubes):
        # XXX Fix the cubes such that the sigma coordinate is always an AuxCoord. Pending gh issue #18
        if isinstance(cubes, iris.cube.Cube):
            cubes = [cubes]

        for cube in cubes:
            sigma = cube.coord("sigma")
            sigma = iris.coords.AuxCoord.from_coord(sigma)
            cube.replace_coord(sigma)

    def assertCML(self, cubes, filename):
        filename = "%s_%s.cml" % (filename, self.suffix)
        tests.IrisTest.assertCML(self, cubes, ("constrained_load", filename))

    def load_match(self, files, constraints):
        raise NotImplementedError()  # defined in subclasses

    def test_single_atomic_constraint(self):
        cubes = self.load_match(self.dec_path, self.level_10)
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "all_10")

        cubes = self.load_match(self.dec_path, self.theta)
        self.assertCML(cubes, "theta")

        cubes = self.load_match(self.dec_path, self.model_level_number_10_22)
        self.fixup_sigma_to_be_aux(cubes)
        workaround_pending_1262(cubes)
        self.assertCML(cubes, "all_ml_10_22")

        # Check that it didn't matter that we provided sets & tuples to the model_level
        for constraint in [
            iris.Constraint(model_level_number=set([10, 22])),
            iris.Constraint(model_level_number=tuple([10, 22])),
        ]:
            cubes = self.load_match(self.dec_path, constraint)
            self.fixup_sigma_to_be_aux(cubes)
            workaround_pending_1262(cubes)
            self.assertCML(cubes, "all_ml_10_22")

    def test_string_standard_name(self):
        cubes = self.load_match(self.dec_path, SN_AIR_POTENTIAL_TEMPERATURE)
        self.assertCML(cubes, "theta")

        cubes = self.load_match(self.dec_path, [SN_AIR_POTENTIAL_TEMPERATURE])
        self.assertCML(cubes, "theta")

        cubes = self.load_match(
            self.dec_path, iris.Constraint(SN_AIR_POTENTIAL_TEMPERATURE)
        )
        self.assertCML(cubes, "theta")

        cubes = self.load_match(
            self.dec_path,
            iris.Constraint(
                SN_AIR_POTENTIAL_TEMPERATURE, model_level_number=10
            ),
        )
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_10")

    def test_latitude_constraint(self):
        cubes = self.load_match(self.theta_path, self.lat_30)
        self.assertCML(cubes, "theta_lat_30")

        cubes = self.load_match(self.theta_path, self.lat_gt_45)
        self.assertCML(cubes, "theta_lat_gt_30")

    def test_single_expression_constraint(self):
        cubes = self.load_match(self.theta_path, self.theta & self.level_10)
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_10")

        cubes = self.load_match(self.theta_path, self.level_10 & self.theta)
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_10")

    def test_dual_atomic_constraint(self):
        cubes = self.load_match(self.dec_path, [self.theta, self.level_10])
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_and_all_10")

    def test_dual_repeated_constraint(self):
        cubes = self.load_match(self.dec_path, [self.theta, self.theta])
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_and_theta")

    def test_dual_expression_constraint(self):
        cubes = self.load_match(
            self.dec_path,
            [self.theta & self.level_10, self.level_gt_30_le_3 & self.theta],
        )
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_10_and_theta_level_gt_30_le_3")

    def test_invalid_constraint(self):
        cubes = self.load_match(self.theta_path, self.pressure_950)
        self.assertCML(cubes, "pressure_950")

        cubes = self.load_match(self.theta_path, self.invalid_inequality)
        self.assertCML(cubes, "invalid_inequality")

    def test_inequality_constraint(self):
        cubes = self.load_match(self.theta_path, self.level_gt_30_le_3)
        self.assertCML(cubes, "theta_gt_30_le_3")


class StrictConstraintMixin(RelaxedConstraintMixin):
    def test_single_atomic_constraint(self):
        cubes = self.load_match(self.theta_path, self.theta)
        self.assertCML(cubes, "theta")

        cubes = self.load_match(self.theta_path, self.level_10)
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_10")

    def test_invalid_constraint(self):
        with self.assertRaises(iris.exceptions.ConstraintMismatchError):
            self.load_match(self.theta_path, self.pressure_950)

    def test_dual_atomic_constraint(self):
        cubes = self.load_match(
            self.dec_path, [self.theta, self.level_10 & self.theta]
        )
        self.fixup_sigma_to_be_aux(cubes)
        self.assertCML(cubes, "theta_and_theta_10")


@tests.skip_data
class TestCubeLoadConstraint(RelaxedConstraintMixin, tests.IrisTest):
    suffix = "load_match"

    def load_match(self, files, constraints):
        cubes = iris.load(files, constraints)
        if not isinstance(cubes, iris.cube.CubeList):
            raise Exception("NOT A CUBE LIST! " + str(type(cubes)))
        return cubes


@tests.skip_data
class TestCubeListConstraint(RelaxedConstraintMixin, tests.IrisTest):
    suffix = "load_match"

    def load_match(self, files, constraints):
        cubes = iris.load(files).extract(constraints)
        if not isinstance(cubes, iris.cube.CubeList):
            raise Exception("NOT A CUBE LIST! " + str(type(cubes)))
        return cubes


@tests.skip_data
class TestCubeListStrictConstraint(StrictConstraintMixin, tests.IrisTest):
    suffix = "load_strict"

    def load_match(self, files, constraints):
        cubes = iris.load(files).extract_cubes(constraints)
        return cubes


@tests.skip_data
class TestCubeExtract__names(TestMixin, tests.IrisTest):
    def setUp(self):
        fname = iris.sample_data_path("atlantic_profiles.nc")
        self.cubes = iris.load(fname)
        TestMixin.setUp(self)
        cube = iris.load_cube(self.theta_path)
        # Expected names...
        self.standard_name = "air_potential_temperature"
        self.long_name = "AIR POTENTIAL TEMPERATURE"
        self.var_name = "apt"
        self.stash = "m01s00i004"
        # Configure missing names...
        cube.long_name = self.long_name
        cube.var_name = self.var_name
        # Add this cube to the mix...
        self.cubes.append(cube)
        self.index = len(self.cubes) - 1

    def test_standard_name(self):
        constraint = iris.Constraint(self.standard_name)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)

    def test_long_name(self):
        constraint = iris.Constraint(self.long_name)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.long_name, self.long_name)

    def test_var_name(self):
        constraint = iris.Constraint(self.var_name)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.var_name, self.var_name)

    def test_stash(self):
        constraint = iris.Constraint(self.stash)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(str(result.attributes["STASH"]), self.stash)

    def test_unknown(self):
        cube = self.cubes[self.index]
        # Clear the cube metadata.
        cube.standard_name = None
        cube.long_name = None
        cube.var_name = None
        cube.attributes = None
        # Extract the unknown cube.
        constraint = iris.Constraint("unknown")
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.name(), "unknown")


@tests.skip_data
class TestCubeExtract__name_constraint(TestMixin, tests.IrisTest):
    def setUp(self):
        fname = iris.sample_data_path("atlantic_profiles.nc")
        self.cubes = iris.load(fname)
        TestMixin.setUp(self)
        cube = iris.load_cube(self.theta_path)
        # Expected names...
        self.standard_name = "air_potential_temperature"
        self.long_name = "air potential temperature"
        self.var_name = "apt"
        self.stash = "m01s00i004"
        # Configure missing names...
        cube.long_name = self.long_name
        cube.var_name = self.var_name
        # Add this cube to the mix...
        self.cubes.append(cube)
        self.index = len(self.cubes) - 1

    def test_standard_name(self):
        # No match.
        constraint = NameConstraint(standard_name="wibble")
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        constraint = NameConstraint(standard_name=self.standard_name)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)

        # Match - callable.
        kwargs = dict(standard_name=lambda item: item.startswith("air_pot"))
        constraint = NameConstraint(**kwargs)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)

    def test_standard_name__None(self):
        cube = self.cubes[self.index]
        cube.standard_name = None
        constraint = NameConstraint(
            standard_name=None, long_name=self.long_name
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertIsNone(result.standard_name)
        self.assertEqual(result.long_name, self.long_name)

    def test_long_name(self):
        # No match.
        constraint = NameConstraint(long_name="wibble")
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        constraint = NameConstraint(long_name=self.long_name)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.long_name, self.long_name)

        # Match - callable.
        kwargs = dict(
            long_name=lambda item: item is not None
            and item.startswith("air pot")
        )
        constraint = NameConstraint(**kwargs)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.long_name, self.long_name)

    def test_long_name__None(self):
        cube = self.cubes[self.index]
        cube.long_name = None
        constraint = NameConstraint(
            standard_name=self.standard_name, long_name=None
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)
        self.assertIsNone(result.long_name)

    def test_var_name(self):
        # No match.
        constraint = NameConstraint(var_name="wibble")
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        constraint = NameConstraint(var_name=self.var_name)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.var_name, self.var_name)

        # Match - callable.
        kwargs = dict(var_name=lambda item: item.startswith("ap"))
        constraint = NameConstraint(**kwargs)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.var_name, self.var_name)

    def test_var_name__None(self):
        cube = self.cubes[self.index]
        cube.var_name = None
        constraint = NameConstraint(
            standard_name=self.standard_name, var_name=None
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)
        self.assertIsNone(result.var_name)

    def test_stash(self):
        # No match.
        constraint = NameConstraint(STASH="m01s00i444")
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        constraint = NameConstraint(STASH=self.stash)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(str(result.attributes["STASH"]), self.stash)

        # Match - callable.
        kwargs = dict(STASH=lambda stash: stash.item == 4)
        constraint = NameConstraint(**kwargs)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)

    def test_stash__None(self):
        cube = self.cubes[self.index]
        del cube.attributes["STASH"]
        constraint = NameConstraint(
            standard_name=self.standard_name, STASH=None
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)
        self.assertIsNone(result.attributes.get("STASH"))

    def test_compound(self):
        # Match.
        constraint = NameConstraint(
            standard_name=self.standard_name, long_name=self.long_name
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)

        # No match - var_name.
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name="wibble",
        )
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)
        self.assertEqual(result.long_name, self.long_name)
        self.assertEqual(result.var_name, self.var_name)

        # No match - STASH.
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            STASH="m01s00i444",
        )
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            STASH=self.stash,
        )
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertEqual(result.standard_name, self.standard_name)
        self.assertEqual(result.long_name, self.long_name)
        self.assertEqual(result.var_name, self.var_name)
        self.assertEqual(result.var_name, self.var_name)

        # No match - standard_name.
        constraint = NameConstraint(
            standard_name="wibble",
            long_name=self.long_name,
            var_name=self.var_name,
            STASH=self.stash,
        )
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

    def test_unknown(self):
        # No match.
        constraint = NameConstraint(None, None, None, None)
        result = self.cubes.extract(constraint)
        self.assertFalse(result)

        # Match.
        cube = self.cubes[self.index]
        cube.standard_name = None
        cube.long_name = None
        cube.var_name = None
        cube.attributes = None
        constraint = NameConstraint(None, None, None, None)
        result = self.cubes.extract_cube(constraint)
        self.assertIsNotNone(result)
        self.assertIsNone(result.standard_name)
        self.assertIsNone(result.long_name)
        self.assertIsNone(result.var_name)
        self.assertIsNone(result.attributes.get("STASH"))


@tests.skip_data
class TestCubeExtract(TestMixin, tests.IrisTest):
    def setUp(self):
        TestMixin.setUp(self)
        self.cube = iris.load_cube(self.theta_path)

    def test_attribute_constraint(self):
        # There is no my_attribute on the cube, so ensure it returns None.
        constraint = AttributeConstraint(my_attribute="foobar")
        cube = self.cube.extract(constraint)
        self.assertIsNone(cube)

        orig_cube = self.cube
        # add an attribute to the cubes
        orig_cube.attributes["my_attribute"] = "foobar"

        constraint = AttributeConstraint(my_attribute="foobar")
        cube = orig_cube.extract(constraint)
        self.assertCML(cube, ("constrained_load", "attribute_constraint.cml"))

        constraint = AttributeConstraint(my_attribute="not me")
        cube = orig_cube.extract(constraint)
        self.assertIsNone(cube)

        kwargs = dict(my_attribute=lambda val: val.startswith("foo"))
        constraint = AttributeConstraint(**kwargs)
        cube = orig_cube.extract(constraint)
        self.assertCML(cube, ("constrained_load", "attribute_constraint.cml"))

        kwargs = dict(my_attribute=lambda val: not val.startswith("foo"))
        constraint = AttributeConstraint(**kwargs)
        cube = orig_cube.extract(constraint)
        self.assertIsNone(cube)

        kwargs = dict(my_non_existant_attribute="hello world")
        constraint = AttributeConstraint(**kwargs)
        cube = orig_cube.extract(constraint)
        self.assertIsNone(cube)

    def test_standard_name(self):
        r = iris.Constraint(SN_AIR_POTENTIAL_TEMPERATURE)
        self.assertTrue(
            self.cube.extract(r).standard_name, SN_AIR_POTENTIAL_TEMPERATURE
        )

        r = iris.Constraint("wibble")
        self.assertEqual(self.cube.extract(r), None)

    def test_empty_data(self):
        # Ensure that the process of WHERE does not load data if there
        # was empty data to start with...
        cube = self.cube
        self.assertTrue(cube.has_lazy_data())
        cube = self.cube.extract(self.level_10)
        self.assertTrue(cube.has_lazy_data())
        cube = self.cube.extract(self.level_10).extract(self.level_10)
        self.assertTrue(cube.has_lazy_data())

    def test_non_existent_coordinate(self):
        # Check the behaviour when a constraint is given for a coordinate which does not exist/span a dimension
        self.assertEqual(self.cube[0, :, :].extract(self.level_10), None)

        self.assertEqual(self.cube.extract(iris.Constraint(wibble=10)), None)


@tests.skip_data
class TestConstraints(TestMixin, tests.IrisTest):
    def test_constraint_expressions(self):
        rt = repr(self.theta)
        rl10 = repr(self.level_10)

        rt_l10 = repr(self.theta & self.level_10)
        expr = "ConstraintCombination(%s, %s, <built-in function %s>)" % (
            rt,
            rl10,
            "and_",
        )
        self.assertEqual(expr, rt_l10)

    def test_string_repr(self):
        rt = repr(iris.Constraint(SN_AIR_POTENTIAL_TEMPERATURE))
        self.assertEqual(
            rt, "Constraint(name='%s')" % SN_AIR_POTENTIAL_TEMPERATURE
        )

        rt = repr(
            iris.Constraint(
                SN_AIR_POTENTIAL_TEMPERATURE, model_level_number=10
            )
        )
        self.assertEqual(
            rt,
            "Constraint(name='%s', coord_values={'model_level_number': 10})"
            % SN_AIR_POTENTIAL_TEMPERATURE,
        )

    def test_number_of_raw_cubes(self):
        # Test the constraints generate the correct number of raw cubes.
        raw_cubes = iris.load_raw(self.theta_path)
        self.assertEqual(len(raw_cubes), 38)

        raw_cubes = iris.load_raw(self.theta_path, [self.level_10])
        self.assertEqual(len(raw_cubes), 1)

        raw_cubes = iris.load_raw(self.theta_path, [self.theta])
        self.assertEqual(len(raw_cubes), 38)

        raw_cubes = iris.load_raw(self.dec_path, [self.level_30])
        self.assertEqual(len(raw_cubes), 4)

        raw_cubes = iris.load_raw(self.dec_path, [self.theta])
        self.assertEqual(len(raw_cubes), 38)


class TestBetween(tests.IrisTest):
    def run_test(self, function, numbers, results):
        for number, result in zip(numbers, results):
            self.assertEqual(function(number), result)

    def test_le_ge(self):
        function = iris.util.between(2, 4)
        numbers = [1, 2, 3, 4, 5]
        results = [False, True, True, True, False]
        self.run_test(function, numbers, results)

    def test_lt_gt(self):
        function = iris.util.between(
            2, 4, rh_inclusive=False, lh_inclusive=False
        )
        numbers = [1, 2, 3, 4, 5]
        results = [False, False, True, False, False]
        self.run_test(function, numbers, results)

    def test_le_gt(self):
        function = iris.util.between(2, 4, rh_inclusive=False)
        numbers = [1, 2, 3, 4, 5]
        results = [False, True, True, False, False]
        self.run_test(function, numbers, results)

    def test_lt_ge(self):
        function = iris.util.between(2, 4, lh_inclusive=False)
        numbers = [1, 2, 3, 4, 5]
        results = [False, False, True, True, False]
        self.run_test(function, numbers, results)


if __name__ == "__main__":
    tests.main()
